{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Parametric Box-Wing Inversion\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import datetime\nfrom typing import Any\n\nimport matplotlib.pyplot as plt\nimport numpy as np\nfrom matplotlib.widgets import Button, Slider\nfrom scipy.io import loadmat\nfrom scipy.optimize import minimize\n\nimport mirage as mr\nimport mirage.vis as mrv\n\nvars_enum = [\n    'x_scale',\n    'y_scale',\n    'z_scale',\n    'wing_area',\n    'cd_wing',\n    'cd_box',\n    'cs_box',\n    'cs_wing',\n    'n_box',\n    'n_wing',\n]\n\n\ndef construct_from_y(attitude, y: np.ndarray, knowns: dict):\n    for k, v in knowns.items():\n        y[vars_enum.index(k)] = v\n    cd_wing = np.clip(y[4], 0, 1)\n    cd_box = np.clip(y[5], 0, 1)\n    cs_box = np.clip(y[6], 0, 1)\n    cs_wing = np.clip(y[7], 0, 1)\n    y = np.clip(y, 1e-8, np.inf)\n    n_box, n_wing = y[8], y[9]\n    brdf_box = mr.Brdf(name=brdf_name, cd=cd_box, cs=cs_box, n=n_box, validate=False)\n    brdf_wing = mr.Brdf(\n        name=brdf_name, cd=cd_wing, cs=cs_wing, n=n_wing, validate=False\n    )\n    return BoxWingParametric(\n        attitude,\n        brdf_box,\n        brdf_wing,\n        x_scale=y[0],\n        y_scale=y[1],\n        z_scale=y[2],\n        wing_area=y[3],\n    )\n\n\nclass BoxWingParametric:\n    def __init__(\n        self,\n        attitude: mr.AlignedAndConstrainedAttitude,\n        brdf_box: mr.Brdf,\n        brdf_wing: mr.Brdf,\n        x_scale: float = 1.0,\n        y_scale: float = 1.0,\n        z_scale: float = 1.0,\n        wing_area: float = 1.0,\n    ):\n        self.attitude = attitude\n        self.brdf_box = brdf_box\n        self.brdf_wing = brdf_wing\n        self._cube_template = mr.SpaceObject('cube.obj')\n        self._cube_template.v /= np.sqrt(self._cube_template.unique_areas[0])\n        self.box = mr.SpaceObject(\n            vertices_and_faces=(self._cube_template.v, self._cube_template.f)\n        )\n        self.scales = [x_scale, y_scale, z_scale]\n        self.wing_area = wing_area\n\n    @property\n    def scales(self):\n        return self._scales\n\n    @scales.setter\n    def scales(self, scales):\n        self._scales = scales\n        self.box.v = self._cube_template.v.copy()\n        self.box.v[:, 0] = self.box.v[:, 0] * scales[0]\n        self.box.v[:, 1] = self.box.v[:, 1] * scales[1]\n        self.box.v[:, 2] = self.box.v[:, 2] * scales[2]\n        self.box = mr.SpaceObject(vertices_and_faces=(self.box.v, self.box.f))\n\n    def eval(\n        self,\n        dates: np.ndarray[datetime.datetime, Any],\n        ovi: np.ndarray,\n        jds: np.ndarray = None,\n        eci_to_body: np.ndarray = None,\n        vc_eci: np.ndarray = None,\n    ) -> np.ndarray:\n        if jds is None:\n            jds = mr.date_to_jd(dates)\n        if eci_to_body is None:\n            eci_to_body = self.attitude.dcms_at_dates(dates)\n        if vc_eci is None:\n            vc_eci = self.attitude.const_interpolator(jds)\n        ovb = mr.stack_mat_mult_vec(eci_to_body, ovi)\n        vc_body = mr.stack_mat_mult_vec(eci_to_body, vc_eci)\n\n        box_lc = self.box.convex_light_curve(self.brdf_box, svb=vc_body, ovb=ovb)\n        wing_lc = (\n            self.wing_area\n            * self.brdf_wing.eval_normalized_brightness(\n                L=vc_body, O=ovb, N=vc_body\n            ).flatten()\n        )\n        box_lc *= count_rate * integration_time_s / (rmag * 1e3) ** 2 * isun\n        wing_lc *= count_rate * integration_time_s / (rmag * 1e3) ** 2 * isun\n\n        return box_lc, wing_lc\n\n    def optimize(\n        self,\n        dates: np.ndarray[datetime.datetime, Any],\n        ovi: np.ndarray,\n        lc_unit_observed: np.ndarray,\n        knowns: dict = None,\n    ):\n        if knowns is None:\n            knowns = dict()\n        if hasattr(lc_unit_observed, 'mask'):\n            valid_inds = ~lc_unit_observed.mask\n        else:\n            valid_inds = ~np.isnan(lc_unit_observed)\n        # valid_inds &= (data_index < 350) | (data_index > 500)\n\n        jds = mr.date_to_jd(dates)\n        eci_to_body = self.attitude.dcms_at_dates(dates)\n        vc_eci = self.attitude.const_interpolator(jds)\n\n        def objective(y: np.ndarray) -> float:\n            bwpi = construct_from_y(self.attitude, y, knowns)\n            box_lc, wing_lc = bwpi.eval(dates, ovi, jds, eci_to_body, vc_eci)\n            err = np.linalg.norm(\n                lc_unit_observed[valid_inds] - (box_lc + wing_lc)[valid_inds]\n            )\n            return err\n\n        opt_sol = minimize(objective, y_from_bwp(self), options={'maxiter': 1})\n        return construct_from_y(self.attitude, opt_sol.x, knowns)\n\n    def __repr__(self):\n        return f'BoxWingParametric(attitude={self.attitude}, brdf_box={self.brdf_box}, brdf_wing={self.brdf_wing}, x_scale={self.scales[0]}, y_scale={self.scales[1]}, z_scale={self.scales[2]}, wing_area={self.wing_area})'\n\n\ndef y_from_bwp(bwp: BoxWingParametric):\n    return np.array(\n        [\n            bwp.scales[0],\n            bwp.scales[1],\n            bwp.scales[2],\n            bwp.wing_area,\n            bwp.brdf_wing.cd,\n            bwp.brdf_box.cd,\n            bwp.brdf_box.cs,\n            bwp.brdf_wing.cs,\n            bwp.brdf_box.n,\n            bwp.brdf_wing.n,\n        ]\n    )\n\n\norbit_sol_path = '/Users/liamrobinson/Library/CloudStorage/OneDrive-purdue.edu/pogs/2022/2022-09-18/ProcessedData/Fitted_Orbits/OrbitSolutions.mat'\norbit_mat = loadmat(orbit_sol_path)\nsol_mat = {\n    k: orbit_mat['CuratedObjects'][0][k][0].squeeze()\n    for k in orbit_mat['CuratedObjects'][0].dtype.names\n}\nsol_mat['JD'] = np.sum(sol_mat['JD'], axis=0)\ndates = mr.jd_to_date(sol_mat['JD'])\nrmag = mr.vecnorm(sol_mat['r'].T).flatten()\nisun = mr.total_solar_irradiance_at_dates(dates)\nlc_norm_true = sol_mat['brightness']\n\nepsecs = mr.date_to_epsec(dates)\ndata_index = np.arange(epsecs.size)\n\nstation = mr.Station()\n\nstation.constraints = [\n    mr.SnrConstraint(3),\n    mr.ElevationConstraint(10),\n    mr.TargetIlluminatedConstraint(),\n    mr.ObserverEclipseConstraint(station),\n    mr.VisualMagnitudeConstraint(20),\n    mr.MoonExclusionConstraint(10),\n]\n\nobj = mr.SpaceObject('matlib_gps_iii.obj', identifier='NAVSTAR 80 (USA 309)')\nr_obj_j2k = obj.propagate(dates)\n\nsv = mr.sun(dates)\nnadir = -mr.hat(r_obj_j2k)\nattitude = mr.AlignedAndConstrainedAttitude(\n    v_align=nadir,\n    v_const=sv,\n    dates=dates,\n    axis_order=(2, 0, 1),\n)\n\nvars_enum = [\n    'x_scale',\n    'y_scale',\n    'z_scale',\n    'wing_area',\n    'cd_wing',\n    'cd_box',\n    'cs_box',\n    'cs_wing',\n    'n_box',\n    'n_wing',\n]\n\nbrdf_name = 'cook-torrance'\n\n\nbrdf_wing = mr.Brdf(brdf_name, cd=0.0, cs=0.7, n=0.12)\nbrdf_box = mr.Brdf(brdf_name, cd=0.05, cs=0.23, n=0.05)\n\nstation_pos_eci = station.j2000_at_dates(dates)\nobject_pos_eci = obj.propagate(dates)\novi = mr.hat(station_pos_eci - object_pos_eci)\nz_obs = mr.angle_between_vecs(station_pos_eci, -ovi)\novb = mr.stack_mat_mult_vec(attitude.dcms_at_dates(dates), ovi)\n\nintegration_time_s = 4\nspec = mr.scaled_solar_spectrum(1)\ncount_rate = mr.integrated_spectrum(station, z_obs, spectrum=spec)\n\n\nwing_area = 28.5212  # m^2\nx_scale = 2.4638  # m\ny_scale = 3.4036  # m\nz_scale = 1.778  # m\nknowns = dict(\n    wing_area=wing_area,\n    x_scale=x_scale,\n    y_scale=y_scale,\n    z_scale=z_scale,\n)\n\nbwp = BoxWingParametric(attitude, brdf_box, brdf_wing)\nbox_lc, wing_lc = bwp.eval(dates, ovi)\n\nfig, ax = plt.subplots(figsize=(10, 5))\nplt.subplots_adjust(bottom=0.5)\nblc = plt.plot(data_index, box_lc + wing_lc, label='Fit')\nplt.scatter(data_index, lc_norm_true, label='Observed', s=1)\nmrv.texit('Parametric Box-Wing Fit', 'Date', 'Normalized Brightness')\nplt.legend()\n\n\ndef _update_blc():\n    box_lc, wing_lc = bwp.eval(dates, ovi)\n    blc[0].set_ydata(box_lc + wing_lc)\n    plt.draw()\n\n\ndef update_box_n(val):\n    bwp.brdf_box.n = val\n    _update_blc()\n\n\ndef update_wing_n(val):\n    bwp.brdf_wing.n = val\n    _update_blc()\n\n\ndef update_box_cd(val):\n    bwp.brdf_box.cd = val\n    _update_blc()\n\n\ndef update_wing_cd(val):\n    bwp.brdf_wing.cd = val\n    _update_blc()\n\n\ndef update_box_cs(val):\n    bwp.brdf_box.cs = val\n    _update_blc()\n\n\ndef update_wing_cs(val):\n    bwp.brdf_wing.cs = val\n    _update_blc()\n\n\ndef update_wing_area(val):\n    bwp.wing_area = val\n    _update_blc()\n\n\ndef update_x_scale(val):\n    bwp.scales = [val, bwp.scales[1], bwp.scales[2]]\n    _update_blc()\n\n\ndef update_y_scale(val):\n    bwp.scales = [bwp.scales[0], val, bwp.scales[2]]\n    _update_blc()\n\n\ndef update_z_scale(val):\n    bwp.scales = [bwp.scales[0], bwp.scales[1], val]\n    _update_blc()\n\n\ndy = 0.07\n\nslider_box_cd = Slider(\n    plt.axes([0.1, 0.01, 0.2, 0.03]), 'box_cd', 0, 1, valinit=brdf_box.cd\n)\nslider_box_cd.on_changed(update_box_cd)\n\nslider_box_cs = Slider(\n    plt.axes([0.4, 0.01, 0.2, 0.03]), 'box_cs', 0, 1, valinit=brdf_box.cs\n)\nslider_box_cs.on_changed(update_box_cs)\n\nslider_box_n = Slider(\n    plt.axes([0.7, 0.01, 0.2, 0.03]), 'box_n', 1e-8, 2, valinit=brdf_box.n\n)\nslider_box_n.on_changed(update_box_n)\n\nslider_wing_cd = Slider(\n    plt.axes([0.1, dy, 0.2, 0.03]), 'wing_cd', 0, 1, valinit=brdf_wing.cd\n)\nslider_wing_cd.on_changed(update_wing_cd)\n\nslider_wing_cs = Slider(\n    plt.axes([0.4, dy, 0.2, 0.03]), 'wing_cs', 0, 1, valinit=brdf_wing.cs\n)\nslider_wing_cs.on_changed(update_wing_cs)\n\nslider_wing_n = Slider(\n    plt.axes([0.7, dy, 0.2, 0.03]), 'wing_n', 1e-8, 2, valinit=brdf_wing.n\n)\nslider_wing_n.on_changed(update_wing_n)\n\nslider_x_scale = Slider(\n    plt.axes([0.1, 2 * dy, 0.2, 0.03]), 'x_scale', 1e-8, 5, valinit=x_scale\n)\nslider_x_scale.on_changed(update_x_scale)\n\nslider_y_scale = Slider(\n    plt.axes([0.4, 2 * dy, 0.2, 0.03]), 'y_scale', 1e-8, 5, valinit=y_scale\n)\nslider_y_scale.on_changed(update_y_scale)\n\nslider_z_scale = Slider(\n    plt.axes([0.7, 2 * dy, 0.2, 0.03]), 'z_scale', 1e-8, 5, valinit=z_scale\n)\nslider_z_scale.on_changed(update_z_scale)\n\nslider_wing_area = Slider(\n    plt.axes([0.1, 3 * dy, 0.2, 0.03]), 'wing_area', 1e-8, 50, valinit=wing_area\n)\nslider_wing_area.on_changed(update_wing_area)\n\n\ndef _update_sliders_from_bwp(bwpi: BoxWingParametric):\n    slider_box_cd.set_val(bwpi.brdf_box.cd)\n    slider_box_cs.set_val(bwpi.brdf_box.cs)\n    slider_box_n.set_val(bwpi.brdf_box.n)\n    slider_wing_cd.set_val(bwpi.brdf_wing.cd)\n    slider_wing_cs.set_val(bwpi.brdf_wing.cs)\n    slider_wing_n.set_val(bwpi.brdf_wing.n)\n    slider_x_scale.set_val(bwpi.scales[0])\n    slider_y_scale.set_val(bwpi.scales[1])\n    slider_z_scale.set_val(bwpi.scales[2])\n    slider_wing_area.set_val(bwpi.wing_area)\n    plt.draw()\n\n\ndef _opt_button_callback(*args):\n    global bwp\n    bwp = bwp.optimize(dates, ovi, lc_norm_true, knowns=knowns)\n    _update_sliders_from_bwp(bwp)\n    _update_blc()\n\n\nopt_button = Button(plt.axes([0.4, 3 * dy, 0.2, 0.03]), 'Optimize')\nopt_button.on_clicked(_opt_button_callback)\n_opt_button_callback()\n\n# bwp_reconstructed = construct_from_y(attitude, y_from_bwp(bwp), knowns)\n# print(repr(bwp_reconstructed))\n# print(repr(bwp))\nplt.show()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.11.11"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}